/*
 *  Shaders - Vulkan style
 *
 *  Simplified and refactored from vulkan-tutorial.com
 */

#define GLFW_INCLUDE_VULKAN
#include <GLFW/glfw3.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>

GLFWwindow*      window;                    // GLFW window
VkInstance       instance;                  // Vulkan instance
VkSurfaceKHR     surface;                   // Drawing surface
VkDevice         device;                    // Selected device
VkQueue          graphicsQueue;             // Graphics queue
VkQueue          presentQueue;              // Present queue
uint32_t         swapCount;                 // Swap count
VkSwapchainKHR   swapChain;                 // Swap chain
VkExtent2D       swapChainExtent;           // Swap chain extent
VkImageView*     swapChainImageViews;       // Swap chain image views
VkFramebuffer*   swapChainFramebuffers;     // Swap chain framebuffers
VkRenderPass     renderPass;                // Render pass
VkPipelineLayout pipelineLayout;            // Pipeline layout
VkPipeline       graphicsPipeline;          // Graphics pipeline
VkCommandPool    commandPool;               // Command pool
VkCommandBuffer  commandBuffer;             // Command buffer
VkSemaphore      imageAvailableSemaphore;   // Image available semaphore
VkSemaphore      renderFinishedSemaphore;   // Render finished sempahore
VkFence          inFlightFence;             // In flight fence

//  Declare variable and initialize to zero
#define DECL(type,var) type var;memset(&var,0,sizeof(type))

//
//  Print message to stderr and exit
//
void Fatal(const char* format , ...)
{
   va_list args;
   va_start(args,format);
   vfprintf(stderr,format,args);
   va_end(args);
   exit(1);
}

//
//  Create shader module
//
VkShaderModule createShaderModule(const char* file)
{
   //  Open file
   FILE* f = fopen(file,"rb");
   if (!f) Fatal("Cannot open shader file %s\n",file);
   //  Seek to end to determine size, then rewind
   fseek(f,0,SEEK_END);
   int n = ftell(f);
   rewind(f);
   //  Allocate memory for the whole file
   uint32_t* code = (uint32_t*)malloc(n);
   if (!code) Fatal("Failed to malloc %d bytes for shader file %s\n",n,file);
   //  Snarf and close the file
   if (fread(code,n,1,f)!=1) Fatal("Cannot read %d bytes from shader file %s\n",n,file);
   fclose(f);
   //  Fill in shader struct
   DECL(VkShaderModuleCreateInfo,shaderInfo);
   shaderInfo.sType    = VK_STRUCTURE_TYPE_SHADER_MODULE_CREATE_INFO;
   shaderInfo.codeSize = n;
   shaderInfo.pCode    = code;
   //  Create shader
   VkShaderModule shaderModule;
   if (vkCreateShaderModule(device,&shaderInfo,NULL,&shaderModule))
      Fatal("Failed to create shader module\n");
   //  Free code and return
   free(code);
   return shaderModule;
}

//
//  Initialize vulkan
//
void vulkanInit()
{
   //  Required device extensions
   const char* deviceExtensions[] = {VK_KHR_SWAPCHAIN_EXTENSION_NAME};

   //
   //  Create Vulkan instance
   //
   DECL(VkApplicationInfo,appInfo);
   appInfo.sType              = VK_STRUCTURE_TYPE_APPLICATION_INFO;
   appInfo.pApplicationName   = "Hello Triangle";
   appInfo.applicationVersion = VK_MAKE_VERSION(1,0,0);
   appInfo.pEngineName        = "No Engine";
   appInfo.engineVersion      = VK_MAKE_VERSION(1,0,0);
   appInfo.apiVersion         = VK_API_VERSION_1_0;

   uint32_t glfwExtensionCount = 0;
   const char** glfwExtensions = glfwGetRequiredInstanceExtensions(&glfwExtensionCount);

   DECL(VkInstanceCreateInfo,createInfo);
   createInfo.sType                   = VK_STRUCTURE_TYPE_INSTANCE_CREATE_INFO;
   createInfo.pApplicationInfo        = &appInfo;
   createInfo.enabledExtensionCount   = glfwExtensionCount;
   createInfo.ppEnabledExtensionNames = glfwExtensions;
   if (vkCreateInstance(&createInfo,NULL,&instance))
      Fatal("Failed to create instance\n");

   //  Create window surface
   if (glfwCreateWindowSurface(instance,window,NULL,&surface))
      Fatal("Failed to create window surface\n");

   //
   //  Get available devices
   //
   uint32_t deviceCount=0;
   vkEnumeratePhysicalDevices(instance,&deviceCount,NULL);
   if (deviceCount==0) Fatal("Failed to find GPUs with Vulkan support\n");
   VkPhysicalDevice* devices = (VkPhysicalDevice*)malloc(deviceCount*sizeof(VkPhysicalDevice));
   if (!devices) Fatal("Failed to malloc %d VkPhysicalDevice\n",deviceCount);
   vkEnumeratePhysicalDevices(instance,&deviceCount,devices);

   //
   //  Select phyical device
   //
   VkPhysicalDevice         physicalDevice = VK_NULL_HANDLE;
   VkSurfaceCapabilitiesKHR capabilities;
   uint32_t                 formatCount;
   VkSurfaceFormatKHR*      formats=NULL;
   uint32_t                 presentModeCount;
   VkPresentModeKHR*        presentModes=NULL;
   uint32_t                 graphicsFamily;
   uint32_t                 presentFamily;
   for (int k=0;k<deviceCount && physicalDevice==VK_NULL_HANDLE;k++)
   {
      uint32_t queueFamilyCount = 0;
      vkGetPhysicalDeviceQueueFamilyProperties(devices[k],&queueFamilyCount,NULL);
      VkQueueFamilyProperties* queueFamilies = (VkQueueFamilyProperties*)malloc(queueFamilyCount*sizeof(VkQueueFamilyProperties));
      if (!queueFamilies) Fatal("Failed to malloc %d VkQueueFamilyProperties\n",queueFamilyCount);
      vkGetPhysicalDeviceQueueFamilyProperties(devices[k],&queueFamilyCount,queueFamilies);
      VkBool32 suitable = VK_FALSE;
      for (int i=0;i<queueFamilyCount && !suitable;i++)
      {
         VkBool32 presentSupport = VK_FALSE;
         vkGetPhysicalDeviceSurfaceSupportKHR(devices[k],i,surface,&presentSupport);
         if (presentSupport && queueFamilies[i].queueFlags & VK_QUEUE_GRAPHICS_BIT)
         {
            suitable = VK_TRUE;
            presentFamily  = i;
            graphicsFamily = i;
         }
      }
      free(queueFamilies);

      //  Check that all required device extentions are supported
      if (suitable)
      {
         //  Get device extensions supported
         uint32_t extensionCount;
         vkEnumerateDeviceExtensionProperties(devices[k],NULL,&extensionCount,NULL);
         VkExtensionProperties* availableExtensions = (VkExtensionProperties*)malloc(extensionCount*sizeof(VkExtensionProperties));
         if (!availableExtensions) Fatal("Failed to malloc %d VkExtensionProperties\n",extensionCount);
         vkEnumerateDeviceExtensionProperties(devices[k],NULL,&extensionCount,availableExtensions);

         //  Check if device extentions match
         for (int k=0;k<sizeof(deviceExtensions)/sizeof(char*) && suitable;k++)
         {
            VkBool32 match = VK_FALSE;
            for (int i=0;i<extensionCount && !match;i++)
               if (!strcmp(deviceExtensions[k],availableExtensions[i].extensionName)) match = VK_TRUE;
            if (!match) suitable = VK_FALSE;
         }
         free(availableExtensions);
      }

      //  Get capabilities if device is suitable
      if (suitable)
      {
         //  Get device capabilities
         vkGetPhysicalDeviceSurfaceCapabilitiesKHR(devices[k],surface,&capabilities);
         //  Get formats
         if (formats) free(formats);
         vkGetPhysicalDeviceSurfaceFormatsKHR(devices[k],surface,&formatCount,NULL);
         if (formatCount)
         {
            formats = (VkSurfaceFormatKHR*)malloc(formatCount*sizeof(VkSurfaceFormatKHR));
            if (!formats) Fatal("Failed to malloc %d VkSurfaceFormatKHR\n",formatCount);
            vkGetPhysicalDeviceSurfaceFormatsKHR(devices[k],surface,&formatCount,formats);
         }
         else
            formats = NULL;
         //  Get present modes
         if (presentModes) free(presentModes);
         vkGetPhysicalDeviceSurfacePresentModesKHR(devices[k],surface,&presentModeCount,NULL);
         if (presentModeCount)
         {
            presentModes = (VkPresentModeKHR*)malloc(presentModeCount*sizeof(VkPresentModeKHR));
            if (!presentModes) Fatal("Failed to malloc %d VkPresentModeKHR\n",presentModeCount);
            vkGetPhysicalDeviceSurfacePresentModesKHR(devices[k],surface,&presentModeCount,presentModes);
         }
         else
            presentModes = NULL;
         //  Accept device if formats and presentModes
         if (formats && presentModes)
            physicalDevice = devices[k];
      }
   }
   //  Stop if no GPU is found
   if (physicalDevice==VK_NULL_HANDLE)
      Fatal("Failed to find a suitable GPU\n");

   //
   //  Create logical device
   //
   int      infoCount = (graphicsFamily!=presentFamily) ? 2 : 1;
   float    queuePriority = 1;
   uint32_t uniqueQueueFamilies[] = {graphicsFamily,presentFamily};
   VkDeviceQueueCreateInfo queueCreateInfos[2];
   for (int k=0;k<infoCount;k++)
   {
      DECL(VkDeviceQueueCreateInfo,queueCreateInfo);
      queueCreateInfo.sType            = VK_STRUCTURE_TYPE_DEVICE_QUEUE_CREATE_INFO;
      queueCreateInfo.queueFamilyIndex = uniqueQueueFamilies[k];
      queueCreateInfo.queueCount       = 1;
      queueCreateInfo.pQueuePriorities = &queuePriority;
      queueCreateInfos[k]              = queueCreateInfo;
   }
   DECL(VkPhysicalDeviceFeatures,deviceFeatures);
   DECL(VkDeviceCreateInfo,deviceInfo);
   deviceInfo.sType                   = VK_STRUCTURE_TYPE_DEVICE_CREATE_INFO;
   deviceInfo.queueCreateInfoCount    = infoCount;
   deviceInfo.pQueueCreateInfos       = queueCreateInfos;
   deviceInfo.enabledExtensionCount   = sizeof(deviceExtensions)/sizeof(char*);
   deviceInfo.ppEnabledExtensionNames = deviceExtensions;
   deviceInfo.pEnabledFeatures        = &deviceFeatures;
   if (vkCreateDevice(physicalDevice,&deviceInfo,NULL,&device))
      Fatal("Failed to create logical device %d\n");

   //  Get graphics and present queues
   vkGetDeviceQueue(device,graphicsFamily,0,&graphicsQueue);
   vkGetDeviceQueue(device,presentFamily,0,&presentQueue);

   //  Set surface format
   VkSurfaceFormatKHR surfaceFormat = formats[0];
   for (int k=0;k<formatCount;k++)
      if (formats[k].format==VK_FORMAT_B8G8R8A8_SRGB &&
          formats[k].colorSpace==VK_COLOR_SPACE_SRGB_NONLINEAR_KHR)
         surfaceFormat = formats[k];

   //  Set present mode
   VkPresentModeKHR presentMode = VK_PRESENT_MODE_FIFO_KHR;
   for (int k=0;k<presentModeCount;k++)
      if (presentModes[k]==VK_PRESENT_MODE_MAILBOX_KHR)
         presentMode = presentModes[k];

   //  Set  extent
   VkExtent2D extent;
   if (capabilities.currentExtent.width != UINT32_MAX)
      extent = capabilities.currentExtent;
   else
   {
      int width, height;
      glfwGetFramebufferSize(window,&width,&height);
      if (width<capabilities.minImageExtent.width) width = capabilities.minImageExtent.width;
      if (width>capabilities.maxImageExtent.width) width = capabilities.maxImageExtent.width;
      if (height<capabilities.minImageExtent.height) height = capabilities.minImageExtent.height;
      if (height>capabilities.maxImageExtent.height) height = capabilities.maxImageExtent.height;
      extent.width  = width;
      extent.height = height;
   }

   //  Set images
   uint32_t imageCount = capabilities.minImageCount+1;
   if (capabilities.maxImageCount>0 && imageCount>capabilities.maxImageCount)
      imageCount = capabilities.maxImageCount;

   //  Create swap chain
   DECL(VkSwapchainCreateInfoKHR,chainInfo);
   chainInfo.sType            = VK_STRUCTURE_TYPE_SWAPCHAIN_CREATE_INFO_KHR;
   chainInfo.surface          = surface;
   chainInfo.minImageCount    = imageCount;
   chainInfo.imageFormat      = surfaceFormat.format;
   chainInfo.imageColorSpace  = surfaceFormat.colorSpace;
   chainInfo.imageExtent      = extent;
   chainInfo.imageArrayLayers = 1;
   chainInfo.imageUsage       = VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT;
   uint32_t queueFamilyIndices[] = {graphicsFamily,presentFamily};
   if (graphicsFamily != presentFamily)
   {
      chainInfo.imageSharingMode      = VK_SHARING_MODE_CONCURRENT;
      chainInfo.queueFamilyIndexCount = 2;
      chainInfo.pQueueFamilyIndices   = queueFamilyIndices;
   }
   else
      chainInfo.imageSharingMode = VK_SHARING_MODE_EXCLUSIVE;
   chainInfo.preTransform   = capabilities.currentTransform;
   chainInfo.compositeAlpha = VK_COMPOSITE_ALPHA_OPAQUE_BIT_KHR;
   chainInfo.presentMode    = presentMode;
   chainInfo.clipped        = VK_TRUE;
   chainInfo.oldSwapchain   = VK_NULL_HANDLE;
   if (vkCreateSwapchainKHR(device,&chainInfo,NULL,&swapChain))
      Fatal("Failed to create swap chain\n");

   //  Create swap chain images
   vkGetSwapchainImagesKHR(device,swapChain,&swapCount,NULL);
   VkImage* swapChainImages = (VkImage*)malloc(swapCount*sizeof(VkImage));
   if (!swapChainImages) Fatal("Failed to malloc %d VkImage\n",swapCount);
   vkGetSwapchainImagesKHR(device,swapChain,&swapCount,swapChainImages);
   VkFormat swapChainImageFormat = surfaceFormat.format;
   swapChainExtent = extent;

   //  Create image views
   swapChainImageViews = (VkImageView*)malloc(swapCount*sizeof(VkImageView));
   for (int i=0;i<swapCount;i++)
   {
      DECL(VkImageViewCreateInfo,createInfo);
      createInfo.sType        = VK_STRUCTURE_TYPE_IMAGE_VIEW_CREATE_INFO;
      createInfo.image        = swapChainImages[i];
      createInfo.viewType     = VK_IMAGE_VIEW_TYPE_2D;
      createInfo.format       = swapChainImageFormat;
      createInfo.components.r = VK_COMPONENT_SWIZZLE_IDENTITY;
      createInfo.components.g = VK_COMPONENT_SWIZZLE_IDENTITY;
      createInfo.components.b = VK_COMPONENT_SWIZZLE_IDENTITY;
      createInfo.components.a = VK_COMPONENT_SWIZZLE_IDENTITY;
      createInfo.subresourceRange.aspectMask     = VK_IMAGE_ASPECT_COLOR_BIT;
      createInfo.subresourceRange.baseMipLevel   = 0;
      createInfo.subresourceRange.levelCount     = 1;
      createInfo.subresourceRange.baseArrayLayer = 0;
      createInfo.subresourceRange.layerCount     = 1;
      if (vkCreateImageView(device,&createInfo,NULL,&swapChainImageViews[i]))
         Fatal("Failed to create image views\n");
   }

   //
   //  Create render pass
   //
   DECL(VkAttachmentDescription,colorAttachment);
   colorAttachment.format         = swapChainImageFormat;
   colorAttachment.samples        = VK_SAMPLE_COUNT_1_BIT;
   colorAttachment.loadOp         = VK_ATTACHMENT_LOAD_OP_CLEAR;
   colorAttachment.storeOp        = VK_ATTACHMENT_STORE_OP_STORE;
   colorAttachment.stencilLoadOp  = VK_ATTACHMENT_LOAD_OP_DONT_CARE;
   colorAttachment.stencilStoreOp = VK_ATTACHMENT_STORE_OP_DONT_CARE;
   colorAttachment.initialLayout  = VK_IMAGE_LAYOUT_UNDEFINED;
   colorAttachment.finalLayout    = VK_IMAGE_LAYOUT_PRESENT_SRC_KHR;

   DECL(VkAttachmentReference,colorAttachmentRef);
   colorAttachmentRef.attachment = 0;
   colorAttachmentRef.layout     = VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL;

   DECL(VkSubpassDescription,subpass);
   subpass.pipelineBindPoint    = VK_PIPELINE_BIND_POINT_GRAPHICS;
   subpass.colorAttachmentCount = 1;
   subpass.pColorAttachments    = &colorAttachmentRef;

   DECL(VkSubpassDependency,dependency);
   dependency.srcSubpass    = VK_SUBPASS_EXTERNAL;
   dependency.dstSubpass    = 0;
   dependency.srcStageMask  = VK_PIPELINE_STAGE_COLOR_ATTACHMENT_OUTPUT_BIT;
   dependency.srcAccessMask = 0;
   dependency.dstStageMask  = VK_PIPELINE_STAGE_COLOR_ATTACHMENT_OUTPUT_BIT;
   dependency.dstAccessMask = VK_ACCESS_COLOR_ATTACHMENT_WRITE_BIT;

   DECL(VkRenderPassCreateInfo,renderPassInfo);
   renderPassInfo.sType           = VK_STRUCTURE_TYPE_RENDER_PASS_CREATE_INFO;
   renderPassInfo.attachmentCount = 1;
   renderPassInfo.pAttachments    = &colorAttachment;
   renderPassInfo.subpassCount    = 1;
   renderPassInfo.pSubpasses      = &subpass;
   renderPassInfo.dependencyCount = 1;
   renderPassInfo.pDependencies   = &dependency;
   if (vkCreateRenderPass(device,&renderPassInfo,NULL,&renderPass))
     Fatal("Failed to create render pass\n");

   //
   //  Create graphics pipeline
   //
   //  Vertex shader
   VkShaderModule vertShaderModule = createShaderModule("vert.spv");
   DECL(VkPipelineShaderStageCreateInfo,vertShaderStageInfo);
   vertShaderStageInfo.sType  = VK_STRUCTURE_TYPE_PIPELINE_SHADER_STAGE_CREATE_INFO;
   vertShaderStageInfo.stage  = VK_SHADER_STAGE_VERTEX_BIT;
   vertShaderStageInfo.module = vertShaderModule;
   vertShaderStageInfo.pName  = "main";
   //  Fragment shader
   VkShaderModule fragShaderModule = createShaderModule("frag.spv");
   DECL(VkPipelineShaderStageCreateInfo,fragShaderStageInfo);
   fragShaderStageInfo.sType  = VK_STRUCTURE_TYPE_PIPELINE_SHADER_STAGE_CREATE_INFO;
   fragShaderStageInfo.stage  = VK_SHADER_STAGE_FRAGMENT_BIT;
   fragShaderStageInfo.module = fragShaderModule;
   fragShaderStageInfo.pName  = "main";

   //
   //  Create pipeline
   //
   VkPipelineShaderStageCreateInfo shaderStages[] = {vertShaderStageInfo,fragShaderStageInfo};

   DECL(VkPipelineVertexInputStateCreateInfo,vertexInputInfo);
   vertexInputInfo.sType                           = VK_STRUCTURE_TYPE_PIPELINE_VERTEX_INPUT_STATE_CREATE_INFO;
   vertexInputInfo.vertexBindingDescriptionCount   = 0;
   vertexInputInfo.vertexAttributeDescriptionCount = 0;

   DECL(VkPipelineInputAssemblyStateCreateInfo,inputAssembly);
   inputAssembly.sType                  = VK_STRUCTURE_TYPE_PIPELINE_INPUT_ASSEMBLY_STATE_CREATE_INFO;
   inputAssembly.topology               = VK_PRIMITIVE_TOPOLOGY_TRIANGLE_LIST;
   inputAssembly.primitiveRestartEnable = VK_FALSE;

   DECL(VkPipelineViewportStateCreateInfo,viewportState);
   viewportState.sType         = VK_STRUCTURE_TYPE_PIPELINE_VIEWPORT_STATE_CREATE_INFO;
   viewportState.viewportCount = 1;
   viewportState.scissorCount  = 1;

   DECL(VkPipelineRasterizationStateCreateInfo,rasterizer);
   rasterizer.sType                   = VK_STRUCTURE_TYPE_PIPELINE_RASTERIZATION_STATE_CREATE_INFO;
   rasterizer.depthClampEnable        = VK_FALSE;
   rasterizer.rasterizerDiscardEnable = VK_FALSE;
   rasterizer.polygonMode             = VK_POLYGON_MODE_FILL;
   rasterizer.lineWidth               = 1;
   rasterizer.cullMode                = VK_CULL_MODE_BACK_BIT;
   rasterizer.frontFace               = VK_FRONT_FACE_CLOCKWISE;
   rasterizer.depthBiasEnable         = VK_FALSE;

   DECL(VkPipelineMultisampleStateCreateInfo,multisampling);
   multisampling.sType                = VK_STRUCTURE_TYPE_PIPELINE_MULTISAMPLE_STATE_CREATE_INFO;
   multisampling.sampleShadingEnable  = VK_FALSE;
   multisampling.rasterizationSamples = VK_SAMPLE_COUNT_1_BIT;

   DECL(VkPipelineColorBlendAttachmentState,colorBlendAttachment);
   colorBlendAttachment.colorWriteMask = VK_COLOR_COMPONENT_R_BIT | VK_COLOR_COMPONENT_G_BIT | VK_COLOR_COMPONENT_B_BIT | VK_COLOR_COMPONENT_A_BIT;
   colorBlendAttachment.blendEnable    = VK_FALSE;

   DECL(VkPipelineColorBlendStateCreateInfo,colorBlending);
   colorBlending.sType             = VK_STRUCTURE_TYPE_PIPELINE_COLOR_BLEND_STATE_CREATE_INFO;
   colorBlending.logicOpEnable     = VK_FALSE;
   colorBlending.logicOp           = VK_LOGIC_OP_COPY;
   colorBlending.attachmentCount   = 1;
   colorBlending.pAttachments      = &colorBlendAttachment;
   colorBlending.blendConstants[0] = 0;
   colorBlending.blendConstants[1] = 0;
   colorBlending.blendConstants[2] = 0;
   colorBlending.blendConstants[3] = 0;

   VkDynamicState dynamicStates[] = {VK_DYNAMIC_STATE_VIEWPORT,VK_DYNAMIC_STATE_SCISSOR};
   DECL(VkPipelineDynamicStateCreateInfo,dynamicState);
   dynamicState.sType             = VK_STRUCTURE_TYPE_PIPELINE_DYNAMIC_STATE_CREATE_INFO;
   dynamicState.dynamicStateCount = sizeof(dynamicStates)/sizeof(VkDynamicState);
   dynamicState.pDynamicStates    = dynamicStates;

   DECL(VkPipelineLayoutCreateInfo,pipelineLayoutInfo);
   pipelineLayoutInfo.sType                  = VK_STRUCTURE_TYPE_PIPELINE_LAYOUT_CREATE_INFO;
   pipelineLayoutInfo.setLayoutCount         = 0;
   pipelineLayoutInfo.pushConstantRangeCount = 0;

   if (vkCreatePipelineLayout(device,&pipelineLayoutInfo,NULL,&pipelineLayout))
      Fatal("Failed to create pipeline layout\n");

   DECL(VkGraphicsPipelineCreateInfo,pipelineInfo);
   pipelineInfo.sType = VK_STRUCTURE_TYPE_GRAPHICS_PIPELINE_CREATE_INFO;
   pipelineInfo.stageCount = 2;
   pipelineInfo.pStages = shaderStages;
   pipelineInfo.pVertexInputState = &vertexInputInfo;
   pipelineInfo.pInputAssemblyState = &inputAssembly;
   pipelineInfo.pViewportState      = &viewportState;
   pipelineInfo.pRasterizationState = &rasterizer;
   pipelineInfo.pMultisampleState   = &multisampling;
   pipelineInfo.pColorBlendState    = &colorBlending;
   pipelineInfo.pDynamicState       = &dynamicState;
   pipelineInfo.layout              = pipelineLayout;
   pipelineInfo.renderPass          = renderPass;
   pipelineInfo.subpass             = 0;
   pipelineInfo.basePipelineHandle  = VK_NULL_HANDLE;
   if (vkCreateGraphicsPipelines(device,VK_NULL_HANDLE,1,&pipelineInfo,NULL,&graphicsPipeline))
      Fatal("failed to create graphics pipeline\n");

   vkDestroyShaderModule(device,fragShaderModule,NULL);
   vkDestroyShaderModule(device,vertShaderModule,NULL);

   //
   //  Create frame buffers
   //
   swapChainFramebuffers = (VkFramebuffer*)malloc(swapCount*sizeof(VkFramebuffer));
   for (int i=0;i<swapCount;i++)
   {
      VkImageView attachments[] = { swapChainImageViews[i] };

      DECL(VkFramebufferCreateInfo,framebufferInfo);
      framebufferInfo.sType = VK_STRUCTURE_TYPE_FRAMEBUFFER_CREATE_INFO;
      framebufferInfo.renderPass = renderPass;
      framebufferInfo.attachmentCount = 1;
      framebufferInfo.pAttachments = attachments;
      framebufferInfo.width = swapChainExtent.width;
      framebufferInfo.height = swapChainExtent.height;
      framebufferInfo.layers = 1;

      if (vkCreateFramebuffer(device,&framebufferInfo,NULL,&swapChainFramebuffers[i]))
         Fatal("Failed to create framebuffer\n");
   }

   //
   //  Create command pool
   //
   DECL(VkCommandPoolCreateInfo,poolInfo);
   poolInfo.sType = VK_STRUCTURE_TYPE_COMMAND_POOL_CREATE_INFO;
   poolInfo.flags = VK_COMMAND_POOL_CREATE_RESET_COMMAND_BUFFER_BIT;
   poolInfo.queueFamilyIndex = graphicsFamily;
   if (vkCreateCommandPool(device,&poolInfo,NULL,&commandPool))
       Fatal("Failed to create command pool\n");

   //
   //  Create command buffer
   //
   DECL(VkCommandBufferAllocateInfo,allocInfo);
   allocInfo.sType = VK_STRUCTURE_TYPE_COMMAND_BUFFER_ALLOCATE_INFO;
   allocInfo.commandPool = commandPool;
   allocInfo.level = VK_COMMAND_BUFFER_LEVEL_PRIMARY;
   allocInfo.commandBufferCount = 1;
   if (vkAllocateCommandBuffers(device, &allocInfo, &commandBuffer))
      Fatal("Failed to allocate command buffers\n");

   //
   //  Create synchronization objects
   //
   DECL(VkSemaphoreCreateInfo,semaphoreInfo);
   semaphoreInfo.sType = VK_STRUCTURE_TYPE_SEMAPHORE_CREATE_INFO;
   DECL(VkFenceCreateInfo,fenceInfo);
   fenceInfo.sType = VK_STRUCTURE_TYPE_FENCE_CREATE_INFO;
   fenceInfo.flags = VK_FENCE_CREATE_SIGNALED_BIT;
   if (vkCreateSemaphore(device,&semaphoreInfo,NULL,&imageAvailableSemaphore) ||
       vkCreateSemaphore(device,&semaphoreInfo,NULL,&renderFinishedSemaphore) ||
       vkCreateFence(device,&fenceInfo,NULL,&inFlightFence))
         Fatal("Failed to create synchronization objects for a frame\n");
}

//
//  Record command buffer
//
void recordCommandBuffer(VkCommandBuffer commandBuffer,uint32_t imageIndex)
{
   VkOffset2D origin = {0,0};

   DECL(VkCommandBufferBeginInfo,beginInfo);
   beginInfo.sType = VK_STRUCTURE_TYPE_COMMAND_BUFFER_BEGIN_INFO;
   if (vkBeginCommandBuffer(commandBuffer, &beginInfo))
      Fatal("Failed to begin recording command buffer\n");

   DECL(VkRenderPassBeginInfo,renderPassInfo);
   renderPassInfo.sType             = VK_STRUCTURE_TYPE_RENDER_PASS_BEGIN_INFO;
   renderPassInfo.renderPass        = renderPass;
   renderPassInfo.framebuffer       = swapChainFramebuffers[imageIndex];
   renderPassInfo.renderArea.offset = origin;
   renderPassInfo.renderArea.extent = swapChainExtent;

   VkClearValue clearColor = {{{0.0f, 0.0f, 0.0f, 1.0f}}};
   renderPassInfo.clearValueCount = 1;
   renderPassInfo.pClearValues    = &clearColor;

   vkCmdBeginRenderPass(commandBuffer,&renderPassInfo,VK_SUBPASS_CONTENTS_INLINE);
   vkCmdBindPipeline(commandBuffer,VK_PIPELINE_BIND_POINT_GRAPHICS,graphicsPipeline);

   DECL(VkViewport,viewport);
   viewport.x        = 0;
   viewport.y        = 0;
   viewport.width    = swapChainExtent.width;
   viewport.height   = swapChainExtent.height;
   viewport.minDepth = 0;
   viewport.maxDepth = 1;
   vkCmdSetViewport(commandBuffer,0,1,&viewport);

   DECL(VkRect2D,scissor);
   scissor.offset = origin;
   scissor.extent = swapChainExtent;
   vkCmdSetScissor(commandBuffer,0,1,&scissor);

   vkCmdDraw(commandBuffer,3,1,0,0);

   vkCmdEndRenderPass(commandBuffer);

   if (vkEndCommandBuffer(commandBuffer)) Fatal("Failed to record command buffer\n");
}

//
//  Draw frame (display)
//
void drawFrame()
{
   
   vkWaitForFences(device,1,&inFlightFence,VK_TRUE,UINT64_MAX);
   vkResetFences(device,1,&inFlightFence);

   uint32_t imageIndex;
   vkAcquireNextImageKHR(device,swapChain,UINT64_MAX,imageAvailableSemaphore,VK_NULL_HANDLE,&imageIndex);

   vkResetCommandBuffer(commandBuffer, /*VkCommandBufferResetFlagBits*/ 0);
   recordCommandBuffer(commandBuffer,imageIndex);

   VkSemaphore waitSemaphores[] = {imageAvailableSemaphore};
   VkPipelineStageFlags waitStages[] = {VK_PIPELINE_STAGE_COLOR_ATTACHMENT_OUTPUT_BIT};
   DECL(VkSubmitInfo,submitInfo);
   submitInfo.sType              = VK_STRUCTURE_TYPE_SUBMIT_INFO;
   submitInfo.waitSemaphoreCount = 1;
   submitInfo.pWaitSemaphores    = waitSemaphores;
   submitInfo.pWaitDstStageMask  = waitStages;
   submitInfo.commandBufferCount = 1;
   submitInfo.pCommandBuffers    = &commandBuffer;

   VkSemaphore signalSemaphores[] = {renderFinishedSemaphore};
   submitInfo.signalSemaphoreCount = 1;
   submitInfo.pSignalSemaphores = signalSemaphores;
   if (vkQueueSubmit(graphicsQueue,1,&submitInfo,inFlightFence))
      Fatal("Failed to submit draw command buffer\n");

   VkSwapchainKHR swapChains[] = {swapChain};
   DECL(VkPresentInfoKHR,presentInfo);
   presentInfo.sType              = VK_STRUCTURE_TYPE_PRESENT_INFO_KHR;
   presentInfo.waitSemaphoreCount = 1;
   presentInfo.pWaitSemaphores    = signalSemaphores;
   presentInfo.swapchainCount     = 1;
   presentInfo.pSwapchains        = swapChains;
   presentInfo.pImageIndices      = &imageIndex;
   vkQueuePresentKHR(presentQueue, &presentInfo);
}

//
//  Close vulkan
//
void vulkanClose()
{
   //  Clean up semaphores and commoands
   vkDeviceWaitIdle(device);
   vkDestroySemaphore(device,renderFinishedSemaphore,NULL);
   vkDestroySemaphore(device,imageAvailableSemaphore,NULL);
   vkDestroyFence(device,inFlightFence,NULL);
   vkDestroyCommandPool(device,commandPool,NULL);

   //  Close framebuffers
   for (int k=0;k<swapCount;k++)
      vkDestroyFramebuffer(device,swapChainFramebuffers[k],NULL);
   free(swapChainFramebuffers);

   //  Close graphics pipeline and images
   vkDestroyPipeline(device,graphicsPipeline,NULL);
   vkDestroyPipelineLayout(device,pipelineLayout,NULL);
   vkDestroyRenderPass(device,renderPass,NULL);
   for (int k=0;k<swapCount;k++)
      vkDestroyImageView(device,swapChainImageViews[k],NULL);
   free(swapChainImageViews);

   //  Close swapchain, surface and main instance
   vkDestroySwapchainKHR(device,swapChain,NULL);
   vkDestroyDevice(device,NULL);
   vkDestroySurfaceKHR(instance,surface,NULL);
   vkDestroyInstance(instance,NULL);
}

//
//  Main program
//
int main()
{
   //  Initialize GLFW
   glfwInit();
   glfwWindowHint(GLFW_CLIENT_API,GLFW_NO_API);
   glfwWindowHint(GLFW_RESIZABLE,GLFW_FALSE);
   window = glfwCreateWindow(800,600,"Vulkan",NULL,NULL);

   //  Initialize Vulkan
   vulkanInit();

   //  Main loop
   while (!glfwWindowShouldClose(window))
   {
      glfwPollEvents();
      drawFrame();
   }
 
   //  Cleanup vulkan and GLFW
   vulkanClose();
   glfwDestroyWindow(window);
   glfwTerminate();
   return 0;
}
